# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'main.ui'
##
## Created by: Qt User Interface Compiler version 6.10.1
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QApplication, QComboBox, QCommandLinkButton, QFrame,
    QGridLayout, QGroupBox, QHBoxLayout, QLabel,
    QLineEdit, QMainWindow, QMenu, QMenuBar,
    QPushButton, QSizePolicy, QSpinBox, QStackedWidget,
    QStatusBar, QVBoxLayout, QWidget)

from nton.gui.widgets import FileDropper
import nton.gui.resources.icons_rc

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(420, 547)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(MainWindow.sizePolicy().hasHeightForWidth())
        MainWindow.setSizePolicy(sizePolicy)
        MainWindow.setMinimumSize(QSize(420, 547))
        MainWindow.setMaximumSize(QSize(420, 547))
        icon1 = QIcon()
        icon1.addFile(u":/branding/images/nton.ico", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        MainWindow.setWindowIcon(icon1)
        self.actionOpen = QAction(MainWindow)
        self.actionOpen.setObjectName(u"actionOpen")
        icon2 = QIcon()
        icon2.addFile(u":/menubar/open.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionOpen.setIcon(icon2)
        self.actionClose = QAction(MainWindow)
        self.actionClose.setObjectName(u"actionClose")
        icon3 = QIcon()
        icon3.addFile(u":/menubar/close.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionClose.setIcon(icon3)
        self.actionExit = QAction(MainWindow)
        self.actionExit.setObjectName(u"actionExit")
        icon4 = QIcon()
        icon4.addFile(u":/menubar/exit.png", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionExit.setIcon(icon4)
        self.actionExit.setMenuRole(QAction.QuitRole)
        self.actionTroubleshooting = QAction(MainWindow)
        self.actionTroubleshooting.setObjectName(u"actionTroubleshooting")
        self.actionAbout = QAction(MainWindow)
        self.actionAbout.setObjectName(u"actionAbout")
        self.actionAbout.setMenuRole(QAction.AboutRole)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.centralwidget.sizePolicy().hasHeightForWidth())
        self.centralwidget.setSizePolicy(sizePolicy1)
        self.verticalLayout = QVBoxLayout(self.centralwidget)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(0, 0, 0, 0)
        self.appPage = QStackedWidget(self.centralwidget)
        self.appPage.setObjectName(u"appPage")
        self.page1 = QWidget()
        self.page1.setObjectName(u"page1")
        self.verticalLayout_8 = QVBoxLayout(self.page1)
        self.verticalLayout_8.setObjectName(u"verticalLayout_8")
        self.openLabel = QLabel(self.page1)
        self.openLabel.setObjectName(u"openLabel")
        font = QFont()
        font.setFamilies([u"Segoe UI Light"])
        font.setPointSize(16)
        self.openLabel.setFont(font)
        self.openLabel.setStyleSheet(u"border: 3px solid #dcdcdc;\n"
"            border-style: dashed;\n"
"            border-radius: 8px;")
        self.openLabel.setAlignment(Qt.AlignCenter)

        self.verticalLayout_8.addWidget(self.openLabel)

        self.appPage.addWidget(self.page1)
        self.page2 = QWidget()
        self.page2.setObjectName(u"page2")
        self.verticalLayout_2 = QVBoxLayout(self.page2)
        self.verticalLayout_2.setObjectName(u"verticalLayout_2")
        self.a_nameAuthorIconLayout = QHBoxLayout()
        self.a_nameAuthorIconLayout.setObjectName(u"a_nameAuthorIconLayout")
        self.verticalLayout_3 = QVBoxLayout()
        self.verticalLayout_3.setSpacing(4)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.nameBox = QGroupBox(self.page2)
        self.nameBox.setObjectName(u"nameBox")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(0)
        sizePolicy2.setHeightForWidth(self.nameBox.sizePolicy().hasHeightForWidth())
        self.nameBox.setSizePolicy(sizePolicy2)
        self.horizontalLayout_3 = QHBoxLayout(self.nameBox)
        self.horizontalLayout_3.setObjectName(u"horizontalLayout_3")
        self.name = QLineEdit(self.nameBox)
        self.name.setObjectName(u"name")
        self.name.setMaxLength(512)

        self.horizontalLayout_3.addWidget(self.name)


        self.verticalLayout_3.addWidget(self.nameBox)

        self.authorBox = QGroupBox(self.page2)
        self.authorBox.setObjectName(u"authorBox")
        sizePolicy2.setHeightForWidth(self.authorBox.sizePolicy().hasHeightForWidth())
        self.authorBox.setSizePolicy(sizePolicy2)
        self.horizontalLayout_4 = QHBoxLayout(self.authorBox)
        self.horizontalLayout_4.setObjectName(u"horizontalLayout_4")
        self.author = QLineEdit(self.authorBox)
        self.author.setObjectName(u"author")
        self.author.setMaxLength(256)

        self.horizontalLayout_4.addWidget(self.author)


        self.verticalLayout_3.addWidget(self.authorBox)


        self.a_nameAuthorIconLayout.addLayout(self.verticalLayout_3)

        self.icon = QLabel(self.page2)
        self.icon.setObjectName(u"icon")
        self.icon.setMinimumSize(QSize(125, 125))
        self.icon.setMaximumSize(QSize(125, 125))
        self.icon.setFrameShape(QFrame.Box)
        self.icon.setPixmap(QPixmap(u":/branding/images/sad.png"))
        self.icon.setScaledContents(True)

        self.a_nameAuthorIconLayout.addWidget(self.icon)


        self.verticalLayout_2.addLayout(self.a_nameAuthorIconLayout)

        self.b_generalInfoBox = QGroupBox(self.page2)
        self.b_generalInfoBox.setObjectName(u"b_generalInfoBox")
        sizePolicy2.setHeightForWidth(self.b_generalInfoBox.sizePolicy().hasHeightForWidth())
        self.b_generalInfoBox.setSizePolicy(sizePolicy2)
        self.verticalLayout_5 = QVBoxLayout(self.b_generalInfoBox)
        self.verticalLayout_5.setObjectName(u"verticalLayout_5")
        self.gridLayout_2 = QGridLayout()
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.screenshots = QComboBox(self.b_generalInfoBox)
        self.screenshots.addItem("")
        self.screenshots.addItem("")
        self.screenshots.setObjectName(u"screenshots")

        self.gridLayout_2.addWidget(self.screenshots, 1, 3, 1, 1)

        self.screenshotsLabel = QLabel(self.b_generalInfoBox)
        self.screenshotsLabel.setObjectName(u"screenshotsLabel")

        self.gridLayout_2.addWidget(self.screenshotsLabel, 1, 2, 1, 1)

        self.videoCaptureLabel = QLabel(self.b_generalInfoBox)
        self.videoCaptureLabel.setObjectName(u"videoCaptureLabel")

        self.gridLayout_2.addWidget(self.videoCaptureLabel, 1, 0, 1, 1)

        self.displayVersionLabel = QLabel(self.b_generalInfoBox)
        self.displayVersionLabel.setObjectName(u"displayVersionLabel")

        self.gridLayout_2.addWidget(self.displayVersionLabel, 0, 2, 1, 1)

        self.displayVersion = QLineEdit(self.b_generalInfoBox)
        self.displayVersion.setObjectName(u"displayVersion")
        self.displayVersion.setMaxLength(16)

        self.gridLayout_2.addWidget(self.displayVersion, 0, 3, 1, 1)

        self.versionLabel = QLabel(self.b_generalInfoBox)
        self.versionLabel.setObjectName(u"versionLabel")

        self.gridLayout_2.addWidget(self.versionLabel, 0, 0, 1, 1)

        self.videoCapture = QComboBox(self.b_generalInfoBox)
        self.videoCapture.addItem("")
        self.videoCapture.addItem("")
        self.videoCapture.addItem("")
        self.videoCapture.setObjectName(u"videoCapture")

        self.gridLayout_2.addWidget(self.videoCapture, 1, 1, 1, 1)

        self.version = QSpinBox(self.b_generalInfoBox)
        self.version.setObjectName(u"version")
        self.version.setEnabled(False)
        self.version.setMaximum(999999999)

        self.gridLayout_2.addWidget(self.version, 0, 1, 1, 1)


        self.verticalLayout_5.addLayout(self.gridLayout_2)


        self.verticalLayout_2.addWidget(self.b_generalInfoBox)

        self.d_titleIdBox = QGroupBox(self.page2)
        self.d_titleIdBox.setObjectName(u"d_titleIdBox")
        sizePolicy2.setHeightForWidth(self.d_titleIdBox.sizePolicy().hasHeightForWidth())
        self.d_titleIdBox.setSizePolicy(sizePolicy2)
        self.verticalLayout_6 = QVBoxLayout(self.d_titleIdBox)
        self.verticalLayout_6.setObjectName(u"verticalLayout_6")
        self.horizontalLayout = QHBoxLayout()
        self.horizontalLayout.setObjectName(u"horizontalLayout")
        self.titleId = QLineEdit(self.d_titleIdBox)
        self.titleId.setObjectName(u"titleId")
        font1 = QFont()
        font1.setFamilies([u"Courier New"])
        self.titleId.setFont(font1)

        self.horizontalLayout.addWidget(self.titleId)

        self.randomizeIdButton = QPushButton(self.d_titleIdBox)
        self.randomizeIdButton.setObjectName(u"randomizeIdButton")

        self.horizontalLayout.addWidget(self.randomizeIdButton)


        self.verticalLayout_6.addLayout(self.horizontalLayout)

        self.titleIdWarning = QLabel(self.d_titleIdBox)
        self.titleIdWarning.setObjectName(u"titleIdWarning")
        self.titleIdWarning.setWordWrap(True)

        self.verticalLayout_6.addWidget(self.titleIdWarning)


        self.verticalLayout_2.addWidget(self.d_titleIdBox)

        self.argsBox = QGroupBox(self.page2)
        self.argsBox.setObjectName(u"argsBox")
        self.verticalLayout_4 = QVBoxLayout(self.argsBox)
        self.verticalLayout_4.setObjectName(u"verticalLayout_4")
        self.args = QLineEdit(self.argsBox)
        self.args.setObjectName(u"args")

        self.verticalLayout_4.addWidget(self.args)

        self.loadRomButton = QPushButton(self.argsBox)
        self.loadRomButton.setObjectName(u"loadRomButton")

        self.verticalLayout_4.addWidget(self.loadRomButton)


        self.verticalLayout_2.addWidget(self.argsBox)

        self.e_buildLayout = QHBoxLayout()
        self.e_buildLayout.setObjectName(u"e_buildLayout")
        self.buildButton = QCommandLinkButton(self.page2)
        self.buildButton.setObjectName(u"buildButton")
        sizePolicy.setHeightForWidth(self.buildButton.sizePolicy().hasHeightForWidth())
        self.buildButton.setSizePolicy(sizePolicy)
        self.buildButton.setMaximumSize(QSize(182, 41))

        self.e_buildLayout.addWidget(self.buildButton)


        self.verticalLayout_2.addLayout(self.e_buildLayout)

        self.appPage.addWidget(self.page2)

        self.verticalLayout.addWidget(self.appPage)

        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QMenuBar(MainWindow)
        self.menubar.setObjectName(u"menubar")
        self.menubar.setGeometry(QRect(0, 0, 420, 22))
        self.menuFile = QMenu(self.menubar)
        self.menuFile.setObjectName(u"menuFile")
        self.menuHelp = QMenu(self.menubar)
        self.menuHelp.setObjectName(u"menuHelp")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        self.statusbar.setEnabled(True)
        self.statusbar.setSizeGripEnabled(False)
        MainWindow.setStatusBar(self.statusbar)
#if QT_CONFIG(shortcut)
        self.screenshotsLabel.setBuddy(self.screenshots)
        self.videoCaptureLabel.setBuddy(self.videoCapture)
        self.displayVersionLabel.setBuddy(self.displayVersion)
        self.versionLabel.setBuddy(self.version)
#endif // QT_CONFIG(shortcut)
        QWidget.setTabOrder(self.name, self.author)
        QWidget.setTabOrder(self.author, self.version)
        QWidget.setTabOrder(self.version, self.displayVersion)
        QWidget.setTabOrder(self.displayVersion, self.videoCapture)
        QWidget.setTabOrder(self.videoCapture, self.screenshots)
        QWidget.setTabOrder(self.screenshots, self.buildButton)

        self.menubar.addAction(self.menuFile.menuAction())
        self.menubar.addAction(self.menuHelp.menuAction())
        self.menuFile.addAction(self.actionOpen)
        self.menuFile.addAction(self.actionClose)
        self.menuFile.addAction(self.actionExit)
        self.menuHelp.addAction(self.actionTroubleshooting)
        self.menuHelp.addSeparator()
        self.menuHelp.addAction(self.actionAbout)

        self.retranslateUi(MainWindow)

        self.appPage.setCurrentIndex(1)


        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"NTON v1.0.0", None))
        self.actionOpen.setText(QCoreApplication.translate("MainWindow", u"Open", None))
#if QT_CONFIG(shortcut)
        self.actionOpen.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+O", None))
#endif // QT_CONFIG(shortcut)
        self.actionClose.setText(QCoreApplication.translate("MainWindow", u"Close", None))
#if QT_CONFIG(shortcut)
        self.actionClose.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+W", None))
#endif // QT_CONFIG(shortcut)
        self.actionExit.setText(QCoreApplication.translate("MainWindow", u"Exit", None))
#if QT_CONFIG(shortcut)
        self.actionExit.setShortcut(QCoreApplication.translate("MainWindow", u"Alt+F4", None))
#endif // QT_CONFIG(shortcut)
        self.actionTroubleshooting.setText(QCoreApplication.translate("MainWindow", u"Troubleshooting", None))
        self.actionAbout.setText(QCoreApplication.translate("MainWindow", u"About", None))
        self.openLabel.setText(QCoreApplication.translate("MainWindow", u"<html><head/><body><p><span style=\"\n"
"            font-weight:700;\">Choose an NRO</span> or drag it\n"
"            here.</p></body></html>", None))
        self.nameBox.setTitle(QCoreApplication.translate("MainWindow", u"&Name", None))
        self.name.setText(QCoreApplication.translate("MainWindow", u"nx-hbmenu", None))
        self.authorBox.setTitle(QCoreApplication.translate("MainWindow", u"&Author", None))
        self.author.setText(QCoreApplication.translate("MainWindow", u"switchbrew", None))
        self.icon.setText("")
        self.b_generalInfoBox.setTitle(QCoreApplication.translate("MainWindow", u"General Information", None))
        self.screenshots.setItemText(0, QCoreApplication.translate("MainWindow", u"Enabled", None))
        self.screenshots.setItemText(1, QCoreApplication.translate("MainWindow", u"Disabled", None))

        self.screenshotsLabel.setText(QCoreApplication.translate("MainWindow", u"Screenshots", None))
        self.videoCaptureLabel.setText(QCoreApplication.translate("MainWindow", u"Video Capture", None))
        self.displayVersionLabel.setText(QCoreApplication.translate("MainWindow", u"&Display Version", None))
        self.displayVersion.setText(QCoreApplication.translate("MainWindow", u"3.5.1", None))
        self.versionLabel.setText(QCoreApplication.translate("MainWindow", u"&Version", None))
        self.videoCapture.setItemText(0, QCoreApplication.translate("MainWindow", u"Disabled", None))
        self.videoCapture.setItemText(1, QCoreApplication.translate("MainWindow", u"Manual", None))
        self.videoCapture.setItemText(2, QCoreApplication.translate("MainWindow", u"Enabled", None))

#if QT_CONFIG(tooltip)
        self.version.setToolTip(QCoreApplication.translate("MainWindow", u"hacBrewPack cannot yet specify the Application Version, see\n"
"                 https://github.com/The-4n/hacBrewPack/issues/15", None))
#endif // QT_CONFIG(tooltip)
        self.d_titleIdBox.setTitle(QCoreApplication.translate("MainWindow", u"Title ID", None))
        self.titleId.setInputMask(QCoreApplication.translate("MainWindow", u">BBHHHHHHHHHHHBBB", None))
        self.titleId.setText(QCoreApplication.translate("MainWindow", u"010D6FD3B35CD000", None))
        self.randomizeIdButton.setText(QCoreApplication.translate("MainWindow", u"Randomize", None))
        self.titleIdWarning.setText(QCoreApplication.translate("MainWindow", u"Warning: Setting the Title ID to that of another Title will\n"
"               override it, and in Tinfoil's case, without warning. Overriding a\n"
"               System Title will brick your system!", None))
        self.argsBox.setTitle(QCoreApplication.translate("MainWindow", u"Arguments", None))
        self.args.setPlaceholderText(QCoreApplication.translate("MainWindow", u"Extra parameters used when calling the NRO...", None))
        self.loadRomButton.setText(QCoreApplication.translate("MainWindow", u"Create ROM Forwarder", None))
        self.buildButton.setText(QCoreApplication.translate("MainWindow", u"Build NSP Forwarder", None))
#if QT_CONFIG(shortcut)
        self.buildButton.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+Space", None))
#endif // QT_CONFIG(shortcut)
        self.buildButton.setDescription("")
        self.menuFile.setTitle(QCoreApplication.translate("MainWindow", u"&File", None))
        self.menuHelp.setTitle(QCoreApplication.translate("MainWindow", u"Help", None))
    # retranslateUi




# Source: https://switchbrew.org/wiki/Title_list
# Last updated: 2022-11-12 08:00 UTC+0
# When updating make sure all title IDs are lowercase hex, 16 digits in length (a-z0-9{16}).
import time

import jsonpickle
import requests

from nton.constants import Files

system_modules = (
    # https://switchbrew.org/wiki/Title_list#System_Modules
    "0100000000000000",  # https://switchbrew.org/wiki/Filesystem_services
    "0100000000000001",  # https://switchbrew.org/wiki/Loader_services
    "0100000000000002",  # https://switchbrew.org/wiki/NCM_services
    "0100000000000003",  # https://switchbrew.org/wiki/Process_Manager_services
    "0100000000000004",  # https://switchbrew.org/wiki/Services_API
    "0100000000000005",  # https://switchbrew.org/wiki/Boot
    "0100000000000006",  # https://switchbrew.org/wiki/USB_services
    "0100000000000007",  # https://switchbrew.org/wiki/TMA_services
    "0100000000000008",  # https://switchbrew.org/wiki/Boot2
    "0100000000000009",  # https://switchbrew.org/wiki/Settings_services
    "010000000000000a",  # https://switchbrew.org/wiki/Bus_services
    "010000000000000b",  # https://switchbrew.org/wiki/Bluetooth_Driver_services
    "010000000000000c",  # https://switchbrew.org/wiki/BCAT_services
    "010000000000000d",  # dmnt (currently not present on retail devices)
    "010000000000000e",  # https://switchbrew.org/wiki/Friend_services
    "010000000000000f",  # https://switchbrew.org/wiki/Network_Interface_services
    "0100000000000010",  # https://switchbrew.org/wiki/PTM_services
    "0100000000000011",  # shell (currently not present on retail devices)
    "0100000000000012",  # https://switchbrew.org/wiki/Sockets_services
    "0100000000000013",  # https://switchbrew.org/wiki/HID_services
    "0100000000000014",  # https://switchbrew.org/wiki/Audio_services
    "0100000000000015",  # https://switchbrew.org/wiki/Log_services
    "0100000000000016",  # https://switchbrew.org/wiki/WLAN_services
    "0100000000000017",  # cs (currently not present on retail devices)
    "0100000000000018",  # https://switchbrew.org/wiki/LDN_services
    "0100000000000019",  # https://switchbrew.org/wiki/NV_services
    "010000000000001a",  # https://switchbrew.org/wiki/PCV_services
    "010000000000001b",  # https://switchbrew.org/wiki/PPC_services / https://switchbrew.org/wiki/Capmtp_services
    "010000000000001c",  # https://switchbrew.org/wiki/Nvnflinger_services
    "010000000000001d",  # https://switchbrew.org/wiki/PCIe_services
    "010000000000001e",  # https://switchbrew.org/wiki/Account_services
    "010000000000001f",  # https://switchbrew.org/wiki/NS_Services
    "0100000000000020",  # https://switchbrew.org/wiki/NFC_services
    "0100000000000021",  # https://switchbrew.org/wiki/PSC_services
    "0100000000000022",  # https://switchbrew.org/wiki/Capture_services
    "0100000000000023",  # https://switchbrew.org/wiki/AM_services
    "0100000000000024",  # https://switchbrew.org/wiki/SSL_services
    "0100000000000025",  # https://switchbrew.org/wiki/NIM_services
    "0100000000000026",  # cec (currently not present on retail devices)
    "0100000000000027",  # tspm (currently not present on retail devices)
    "0100000000000028",  # https://switchbrew.org/wiki/SPL_services
    "0100000000000029",  # https://switchbrew.org/wiki/Backlight_services
    "010000000000002a",  # https://switchbrew.org/wiki/BTM_services
    "010000000000002b",  # https://switchbrew.org/wiki/Error_Report_services
    "010000000000002c",  # time (currently not present on retail devices)
    "010000000000002d",  # https://switchbrew.org/wiki/Display_services
    "010000000000002e",  # https://switchbrew.org/wiki/Parental_Control_services
    "010000000000002f",  # https://switchbrew.org/wiki/NPNS_services
    "0100000000000030",  # https://switchbrew.org/wiki/Error_Upload_services
    "0100000000000031",  # arp / https://switchbrew.org/wiki/Glue_services
    "0100000000000032",  # eclct
    "0100000000000033",  # https://switchbrew.org/wiki/ETicket_services
    "0100000000000034",  # https://switchbrew.org/wiki/Fatal_services
    "0100000000000035",  # https://switchbrew.org/wiki/GRC_services
    "0100000000000036",  # https://switchbrew.org/wiki/Creport
    "0100000000000037",  # https://switchbrew.org/wiki/Loader_services#ldr:ro
    "0100000000000038",  # https://switchbrew.org/wiki/Profiler_services (currently not present on retail devices)
    "0100000000000039",  # https://switchbrew.org/wiki/Shared_Database_services
    "010000000000003a",  # https://switchbrew.org/wiki/Migration_services
    "010000000000003b",  # https://switchbrew.org/wiki/JIT_services
    "010000000000003c",  # https://switchbrew.org/wiki/Jpegdec_services
    "010000000000003d",  # https://switchbrew.org/wiki/Safemode
    "010000000000003e",  # https://switchbrew.org/wiki/OLSC_services
    "010000000000003f",  # dt (currently not present on retail devices)
    "0100000000000040",  # nd (currently not present on retail devices)
    "0100000000000041",  # https://switchbrew.org/wiki/NGCT_services
    "0100000000000042",  # https://switchbrew.org/wiki/PGL_services
    "0100000000000043",  # (currently not present on retail devices)
    "0100000000000044",  # (currently not present on retail devices)
    "0100000000000045",  # https://switchbrew.org/wiki/OMM_services
    "0100000000000046",  # https://switchbrew.org/wiki/Ethernet_services
)

system_data_archives = (
    # https://switchbrew.org/wiki/Title_list#System_Data_Archives
    "0100000000000800",  # https://switchbrew.org/wiki/SSL_services#CertStore
    "0100000000000801",  # ErrorMessage
    "0100000000000802",  # MiiModel
    "0100000000000803",  # BrowserDll
    "0100000000000804",  # Help
    "0100000000000805",  # SharedFont
    "0100000000000806",  # NgWord
    "0100000000000807",  # SsidList
    "0100000000000808",  # Dictionary
    "0100000000000809",  # SystemVersion
    "010000000000080a",  # AvatarImage
    "010000000000080b",  # LocalNews
    "010000000000080c",  # Eula (Eura)
    "010000000000080d",  # UrlBlackList
    "010000000000080e",  # TimeZoneBinary
    "010000000000080f",  # CertStoreCruiser (BrowserCertStore)
    "0100000000000810",  # FontNintendoExtension
    "0100000000000811",  # FontStandard
    "0100000000000812",  # FontKorean
    "0100000000000813",  # FontChineseTraditional
    "0100000000000814",  # FontChineseSimple
    "0100000000000815",  # FontBfcpx
    "0100000000000816",  # SystemUpdate
    "0100000000000818",  # FirmwareDebugSettings
    "0100000000000819",  # BootImagePackage
    "010000000000081a",  # BootImagePackageSafe
    "010000000000081b",  # BootImagePackageExFat
    "010000000000081c",  # BootImagePackageExFatSafe
    "010000000000081d",  # FatalMessage
    "010000000000081e",  # ControllerIcon
    "010000000000081f",  # PlatformConfigIcosa
    "0100000000000820",  # PlatformConfigCopper
    "0100000000000821",  # PlatformConfigHoag
    "0100000000000822",  # ControllerFirmware
    "0100000000000823",  # NgWord2
    "0100000000000824",  # PlatformConfigIcosaMariko
    "0100000000000825",  # ApplicationBlackList
    "0100000000000826",  # RebootlessSystemUpdateVersion
    "0100000000000827",  # ContentActionTable
    "0100000000000828",  # FunctionBlackList
    "0100000000000829",  # PlatformConfigCalcio
    "0100000000000830",  # NgWordT
    "0100000000000831",  # PlatformConfigAula
    "0100000000000832",  # Firmware binaries for Aula's dock
)

system_applets = (
    # https://switchbrew.org/wiki/Title_list#System_Applets
    "0100000000001000",  # qlaunch (SystemAppletMenu)
    "0100000000001001",  # auth (LibraryAppletAuth)
    "0100000000001002",  # cabinet (LibraryAppletCabinet)
    "0100000000001003",  # controller (LibraryAppletController)
    "0100000000001004",  # dataErase (LibraryAppletDataErase)
    "0100000000001005",  # error (LibraryAppletError)
    "0100000000001006",  # netConnect (LibraryAppletNetConnect)
    "0100000000001007",  # playerSelect (LibraryAppletPlayerSelect)
    "0100000000001008",  # swkbd (LibraryAppletSwkbd)
    "0100000000001009",  # miiEdit (LibraryAppletMiiEdit)
    "010000000000100a",  # LibAppletWeb (LibraryAppletWeb)
    "010000000000100b",  # LibAppletShop (LibraryAppletShop)
    "010000000000100c",  # overlayDisp (OverlayApplet)
    "010000000000100d",  # photoViewer (LibraryAppletPhotoViewer)
    "010000000000100e",  # set (LibraryAppletSet)
    "010000000000100f",  # LibAppletOff (LibraryAppletOfflineWeb)
    "0100000000001010",  # LibAppletLns (LibraryAppletLoginShare)
    "0100000000001011",  # LibAppletAuth (LibraryAppletWifiWebAuth)
    "0100000000001012",  # starter (DummyStarter)
    "0100000000001013",  # myPage (LibraryAppletMyPage)
    "0100000000001014",  # PlayReport
    "0100000000001015",  # maintenance (MaintenanceMenu)
    "0100000000001016",
    "0100000000001017",
    "0100000000001018",
    "0100000000001019",
    "010000000000101a",  # gift (LibraryAppletGift)
    "010000000000101b",  # DummyECApplet (LibraryAppletDummyShop)
    "010000000000101c",  # userMigration (LibraryAppletUserMigration)
    "010000000000101d",  # EncounterSys (LibraryAppletPreomiaSys)
    "010000000000101e",
    "010000000000101f",
    "0100000000001020",  # story (LibraryAppletStory)
    "0100000000001021",
    "0100000000001023",
    "0100000000001024",
    "0100000000001025",
    "0100000000001026",
    "0100000000001027",
    "0100000000001028",
    "0100000000001029",
    "010000000000102a",
    "010000000000102b",
    "010000000000102c",
    "010000000000102e",
    "010000000000102f",
    "0100000000001030",
    "0100000000001031",
    "0100000000001032",
    "0100000000001033",
    "0100000000001034",
    "0100000000001037",
    "0100000000001038",  # sample (LibraryAppletSample)
    "010000000000103c",
    "010000000000103e",
    "0100000000001fff",  # EndOceanProgramId
)


development_system_applets = (
    # https://switchbrew.org/wiki/Title_list#Development_System_Applets
    "0100000000002000",  # A2BoardFunction
    "0100000000002001",  # A3Wireless
    "0100000000002002",  # C1LcdAndKey (LcdAndKey)
    "0100000000002003",  # C2UsbHpmic (UsbAndHPMicTest)
    "0100000000002004",  # C3Aging (Aging)
    "0100000000002005",  # C4SixAxis (6axisTest)
    "0100000000002006",  # C5Wireless (AssembledWireless)
    "0100000000002007",  # C7FinalCheck (FinalCheck)
    "010000000000203f",  # AutoCapture
    "0100000000002040",  # DevMenuCommandSystem
    "0100000000002041",  # recovery
    "0100000000002042",  # DevMenuSystem
    "0100000000002044",  # HB-TBIntegrationTest
    "010000000000204d",  # BackupSaveData
    "010000000000204e",  # A4BoardCalWriti (BoardCalWriting)
    "0100000000002054",  # RepairSslCertificate
    "0100000000002055",  # GameCardWriter
    "0100000000002056",  # UsbPdTestTool
    "0100000000002057",  # RepairDeletePctl
    "0100000000002058",  # RepairBackup
    "0100000000002059",  # RepairRestore
    "010000000000205a",  # RepairAccountTransfer
    "010000000000205b",  # RepairAutoNetworkUpdater
    "010000000000205c",  # RefurbishReset
    "010000000000205d",  # RepairAssistCup
    "010000000000205e",  # RepairPairingCutter
    "0100000000002064",  # DevMenu
    "0100000000002065",  # DevMenuApp
    "0100000000002066",  # GetGameCardAsicInfo
    "0100000000002068",  # NfpDebugToolSystem
    "0100000000002069",  # AlbumSynchronizer
    "0100000000002071",  # SnapShotDumper
    "0100000000002073",  # DevMenuSystemApp
    "0100000000002099",  # DevOverlayDisp
    "010000000000209a",  # NandVerifier
    "010000000000209b",  # GpuCoreDumper
    "010000000000209c",  # TestApplication (TestApplicationLauncher)
    "010000000000209e",  # HelloWorld
    "01000000000020a0",  # XcieWriter
    "01000000000020a1",  # GpuOverrunNotifier
    "01000000000020c8",  # NfpDebugTool
    "01000000000020ca",  # NoftWriter
    "01000000000020d0",  # BcatSystemDebugTool
    "01000000000020d1",  # DevSafeModeUpdater
    "01000000000020d3",  # ControllerConnectionAnalyzer
    "01000000000020d4",  # DevKitUpdater
    "01000000000020d6",  # RepairTimeReviser
    "01000000000020d7",  # RepairReinitializeFuelGauge
    "01000000000020da",  # RepairAbortMigration
    "01000000000020dc",  # RepairShowDeviceId
    "01000000000020dd",  # RepairSetCycleCountReliability
    "01000000000020e0",  # Interface
    "01000000000020e1",  # AlbumDownloader
    "01000000000020e3",  # FuelGaugeDumper
    "01000000000020e4",  # UnsafeExtract
    "01000000000020e5",  # UnsafeEngrave
    "01000000000020ee",  # BluetoothSettingTool
    "01000000000020f0",  # ApplicationInstallerRomfs (devmenuapp_installer)
    "0100000000002100",  # DevMenuLotcheckDownloader
    "0100000000002101",  # DevMenuCommand
    "0100000000002102",  # ExportPartition
    "0100000000002103",  # SystemInitializer (SystemInitializ)
    "0100000000002104",  # SystemUpdaterHostFs
    "0100000000002105",  # WriteToStorage
    "0100000000002106",  # CalWriter (CalWriterManu)
    "0100000000002107",  # SettingsManager
    "0100000000002109",  # testBuildSystemIris
    "010000000000210a",  # SystemUpdater
    "010000000000210b",  # nvnflinger_util
    "010000000000210c",  # ControllerFirmwareUpdater
    "010000000000210d",  # testBuildSystemNintendoWare (Test)
    "0100000000002110",  # TestSaveDataCreator
    "0100000000002111",  # C9LcdSpker
    "0100000000002114",  # RankTurn
    "0100000000002116",  # BleTestTool
    "010000000000211a",  # PreinstallAppWriter
    "010000000000211c",  # ControllerSerialFlashTool
    "010000000000211d",  # ControllerFlashWriter
    "010000000000211e",  # Handling
    "010000000000211f",  # Hid
    "0100000000002120",  # ControllerTestApp
    "0100000000002121",  # HidInspectionTool
    "0100000000002124",  # BatteryCyclesEditor
    "0100000000002125",  # UsbFirmwareUpdater
    "0100000000002126",  # PalmaSerialCodeTool
    "0100000000002127",  # renderdoccmd
    "0100000000002128",  # HidInspectionToolProd
    "010000000000212c",  # ExhibitionMenu
    "010000000000212f",  # ExhibitionSaveData
    "0100000000002130",  # LuciaConverter
    "0100000000002133",  # CalDumper
    "0100000000002134",  # AnalogStickEvaluationTool
    "010000000000216d",  # ExhibitionSaveDataSnapshot (Unofficial name)
    "0100000000002178",  # SecureStartupSettings (Unofficial name)
    "010000000000217d",  # CradleFirmwareUpdater
)


debug_system_modules = (
    # https://switchbrew.org/wiki/Title_list#Debug_System_Modules
    "0100000000003002",  # DummyProcess
    "0100000000003003",  # DebugMonitor0
    "0100000000003004",  # SystemHelloWorld
)


development_system_modules = (
    # https://switchbrew.org/wiki/Title_list#Development_System_Modules
    "010000000000b120",  # nvdbgsvc
    "010000000000b123",  # acc:CORNX
    "010000000000b14a",  # manu
    "010000000000b14b",  # ManuUsbLoopBack
    "010000000000b1b8",  # DevFwdbgHbPackage
    "010000000000b1b9",  # DevFwdbgUsbPackage
    "010000000000b1ba",  # ProdFwdbgPackage
    "010000000000b22a",  # scs
    "010000000000b22b",  # ControllerFirmwareDebug
    "010000000000b240",  # htc
)


bdk_system_modules = (
    # https://switchbrew.org/wiki/Title_list#Bdk_System_Modules
    "010000000000c600",  # BdkSample01
    "010000000000c601",  # BdkSample02
    "010000000000c602",  # BdkSample03
    "010000000000c603",  # BdkSample04
)


micro_system_modules = (
    # https://switchbrew.org/wiki/Title_list#Micro_System_Modules
    "010000000000d609",  # dmnt.gen2
    "010000000000d60a",
    "010000000000d60b",
    "010000000000d60c",
    "010000000000d60d",
    "010000000000d60e",
    "010000000000d610",
    "010000000000d611",
    "010000000000d612",
    "010000000000d613",
    "010000000000d614",
    "010000000000d615",
    "010000000000d616",
    "010000000000d617",
    "010000000000d619",
    "010000000000d623",  # DevServer
    "010000000000d633",
)


system_applications = (
    # https://switchbrew.org/wiki/Title_list#System_Applications
    "01008bb00013c000",  # flog (NES emulator)
    "0100069000078000",  # RetailInteractiveDisplayMenu (DevQuestMenu)
    "010000b003486000",  # AudioUsbMicDebugTool
    "0100458001e04000",  # BcatTestApp01
    "0100f910020f8000",  # BcatTestApp02
    "0100b7d0020fc000",  # BcatTestApp03
    "0100132002100000",  # BcatTestApp04
    "0100935002116000",  # BcatTestApp05
    "0100da4002130000",  # BcatTestApp06
    "0100b0f002104000",  # BcatTestApp07
    "010051e002132000",  # BcatTestApp08
    "01004cb0015c8000",  # BcatTestApp09
    "01009720015ca000",  # BcatTestApp10
    "01002f20015c6000",  # BcatTestApp11
    "0100204001f90000",  # BcatTestApp12
    "0100060001f92000",  # BcatTestApp13
    "0100c26001f94000",  # BcatTestApp14
    "0100462001f96000",  # BcatTestApp15
    "01005c6001f98000",  # BcatTestApp16
    "010070000e3c0000",  # EncounterUsr (LibraryAppletPreomiaUsr)
    "010086000e49c000",  # EncounterUsrDummy (LibraryAppletPreomiaUsrDummy)
    "0100810002d5a000",  # ShopMonitaringTool
    "010023d002b98000",  # DeltaStress
    "010099f00d810000",
)


pre_release_system_applets = (
    # https://switchbrew.org/wiki/Title_list#Pre-release_System_Applets
    "1000000000000001",  # SystemInitializer (SystemInitializ)
    "1000000000000004",  # CalWriter (CalWriterManu)
    "1000000000000005",  # DevMenuCommand
    "1000000000000006",  # SettingsManager
    "1000000000000007",  # DevMenu (debug) / TestApplication (factory)
    "100000000000000b",  # SnapShotDumper
    "100000000000000c",  # SystemUpdater
    "100000000000000e",  # ControllerFirmwareUpdater
)


pre_release_system_modules = (
    # https://switchbrew.org/wiki/Title_list#Pre-release_System_Modules
    "1000000000000201",  # usb
    "1000000000000202",  # tma
    "1000000000000203",  # boot2
    "1000000000000204",  # settings
    "1000000000000205",  # Bus
    "1000000000000206",  # bluetooth
    "1000000000000208",  # DebugMonitor0
    "1000000000000209",  # dmnt
    "100000000000020b",  # nifm
    "100000000000020c",  # ptm
    "100000000000020e",  # bsdsocket
    "100000000000020f",  # hid
    "1000000000000210",  # audio
    "1000000000000212",  # LogManager
    "1000000000000213",  # wlan
    "1000000000000214",  # cs
    "1000000000000215",  # ldn
    "1000000000000216",  # nvservices
    "1000000000000217",  # pcv
    "1000000000000218",  # ppc
    "100000000000021a",  # lbl0
    "100000000000021b",  # nvnflinger
    "100000000000021c",  # pcie
    "100000000000021d",  # account
    "100000000000021e",  # ns
    "100000000000021f",  # nfc
    "1000000000000220",  # psc
    "1000000000000221",  # capsrv
    "1000000000000222",  # am
    "1000000000000223",  # ssl
    "1000000000000224",  # nim
)


# The following are unofficial title IDs I have pre-defined to be used for specific homebrew
# This is to keep a specific title ID throughout each release's bundled prebuilt NSPs
unofficial = {
    "hbmenu": "010d6fd3b35cd000",  # https://github.com/switchbrew/nx-hbmenu
    "aio-switch-updater": "016454da0e765000",  # https://github.com/HamletDuFromage/aio-switch-updater
}


def get_game_title_ids() -> dict:
    """Get a mapping of Game Title IDs -> Game Names from Tinfoil's API."""
    res = requests.get(
        url="https://tinfoil.media/Title/ApiJson/",
        params={
            # is this actually returning a full list?
            "rating_content": "",
            "language": "",
            "category": "",
            "region": "ar,at,au,be,bg,br,ca,ch,cl,cn,co,cy,cz,de,dk,ee,es,fi,fr,gb,gr,hk,hr,hu,"
            "ie,it,jp,kr,lt,lu,lv,mt,mx,nl,no,nz,pe,pl,pt,ro,ru,se,si,sk,us,xx,za,zh",
            "rating": "",
            "_": str(time.time_ns() // 1000000),
        },
        headers={
            "Accept": "application/json, text/javascript, */*; q=0.01",
            "User-Agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:109.0) Gecko/20100101 Firefox/115.0",
            "Origin": "https://tinfoil.io",
            "Referer": "https://tinfoil.io/",
        },
    )

    if not res.ok:
        raise requests.ConnectionError(
            f"Failed to get a list of Game Title IDs from Tinfoil's API, [{res.status_code}]"
        )

    title_ids = {
        title["id"].lower(): title["name"].removeprefix("<a>").removesuffix("</a>") for title in res.json()["data"]
    }

    return title_ids


if Files.game_title_ids.exists():
    game_title_ids = jsonpickle.loads(Files.game_title_ids.read_text("utf8"))
else:
    game_title_ids = {}


ALL_SYSTEM = (
    *system_modules,
    *system_data_archives,
    *system_applets,
    *development_system_applets,
    *debug_system_modules,
    *development_system_modules,
    *bdk_system_modules,
    *micro_system_modules,
    *system_applications,
    *pre_release_system_applets,
    *pre_release_system_modules,
)

ALL = (*ALL_SYSTEM, *list(game_title_ids), *list(unofficial.values()))



from __future__ import annotations

import subprocess
from pathlib import Path

from nton.constants import Binaries, Files

BIN = Binaries.nstool
FILE_TYPES = ("xci", "pfs", "romfs", "nca", "meta", "cnmt", "nso", "nro", "ini", "kip", "nacp", "aset", "cert", "tik")
SUBPROCESS_STARTUP_INFO = subprocess.STARTUPINFO()
SUBPROCESS_STARTUP_INFO.dwFlags |= subprocess.STARTF_USESHOWWINDOW


def verify(path: Path, type_: str) -> str | int | None:
    """
    Verify if the Nintendo Switch file is valid.
    Returns an empty string if valid, an error str or return code otherwise.
    """
    if not BIN or not Path(BIN).is_file():
        raise EnvironmentError("nstool binary was not found. Please ensure nstool is installed and in your PATH.")

    if not path:
        raise ValueError("The path must not be empty.")
    if not isinstance(path, Path):
        raise TypeError("The path variable is not a Path object.")
    if not path.is_file():
        raise FileNotFoundError("The path was not found or is not a file.")

    if not type_:
        raise ValueError("The type must not be empty.")
    if not isinstance(type_, str):
        raise TypeError("The type_ variable is not a str object.")
    type_ = type_.lower()
    if type_ not in FILE_TYPES:
        raise ValueError(f'The type_ "{type_}" is not a valid type. File types: {", ".join(FILE_TYPES)}')

    try:
        subprocess.check_output(
            [BIN, "-k", str(Files.keys.absolute()), "-t", type_, "--verify", str(path.absolute())],
            stderr=subprocess.PIPE,
            startupinfo=SUBPROCESS_STARTUP_INFO,
        )
        return None
    except subprocess.CalledProcessError as e:
        return e.output.decode("utf8").strip() or e.returncode


def get_nacp(asset_path: Path, output_path: Path) -> str | int | None:
    """
    Extract the NACP partition from a Homebrew Asset Blob.

    Parameters:
        asset_path: Homebrew Asset Blob to get the NACP from. E.g., an NRO.
        output_path: Path to save the NACP to.

    Returns an empty string if valid, an error str or return code otherwise.
    """
    if not BIN or not Path(BIN).is_file():
        raise EnvironmentError("nstool binary was not found. Please ensure nstool is installed and in your PATH.")

    try:
        subprocess.check_output(
            [BIN, "-k", str(Files.keys.absolute()), "--nacp", str(output_path.absolute()), str(asset_path.absolute())],
            startupinfo=SUBPROCESS_STARTUP_INFO,
        )
        if not output_path.is_file():
            return "No NACP was extracted from the asset."
        if output_path.stat().st_size <= 2:
            return "An empty NACP was extracted from the asset."
        return None
    except subprocess.CalledProcessError as e:
        return e.output.decode("utf8").strip() or e.returncode


def get_icon(asset_path: Path, output_path: Path) -> str | int | None:
    """
    Extract the Icon partition from a Homebrew Asset Blob.

    Parameters:
        asset_path: Homebrew Asset Blob to get the Icon from. E.g., an NRO.
        output_path: Path to save the Icon to.

    Returns an empty string if valid, an error str or return code otherwise.
    """
    if not BIN or not Path(BIN).is_file():
        raise EnvironmentError("nstool binary was not found. Please ensure nstool is installed and in your PATH.")

    try:
        subprocess.check_output(
            [BIN, "-k", str(Files.keys.absolute()), "--icon", str(output_path.absolute()), str(asset_path.absolute())],
            startupinfo=SUBPROCESS_STARTUP_INFO,
        )
        if not output_path.is_file():
            return "No Icon was extracted from the asset."
        if output_path.stat().st_size <= 2:
            return "An empty Icon was extracted from the asset."
        return None
    except subprocess.CalledProcessError as e:
        return e.output.decode("utf8").strip() or e.returncode



import ctypes
import os
import re
import shutil
import string
import subprocess
import sys
import tempfile
import webbrowser
from functools import partial
from pathlib import Path
from typing import Optional, Union

from PIL import Image
from PySide6.QtGui import QPixmap
from PySide6.QtUiTools import QUiLoader
from PySide6.QtWidgets import QApplication, QFileDialog, QInputDialog, QMainWindow, QMessageBox

from nton import __version__, nstool, title_ids
from nton.constants import Binaries, Directories, Files
from nton.gui.logger import log
from nton.gui.main import Ui_MainWindow
from nton.gui.widgets import FileDropper
from nton.helpers import get_copyright_years

NRO_PATH: Optional[Path]
SDMC: Optional[str]
CONTROL_NACP: bytearray

RE_TITLE_ID = re.compile(r"^01([a-fA-F0-9]{11})000$")


class MainWindow(QMainWindow):
    def __init__(self) -> None:
        super().__init__()
        self.ui = Ui_MainWindow()
        self.ui.setupUi(self)


def start(debug: bool = False) -> None:
    """Start the GUI and Qt execution loop."""
    # https://stackoverflow.com/a/12522799/13183782
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID("com.rlaphoenix.nton")

    if debug:
        log.setLevel("DEBUG")

    gui = QApplication(sys.argv)
    loader = QUiLoader()
    loader.registerCustomWidget(FileDropper)

    window = MainWindow()
    window.setWindowTitle(f"NTON v{__version__}")

    ensure_dependencies(window)
    setup_logic(window)
    reset_ui(window)  # clear UI file's placeholder data

    window.show()

    gui.exec_()


def ensure_dependencies(window: MainWindow) -> None:
    for binary, path in vars(Binaries).items():
        if binary.startswith("__"):
            continue
        if not path:
            QMessageBox.critical(
                window,
                "Corrupt Installation",
                f"Could not find binary for {binary} which should be pre-bundled with NTON.<br/><br/>"
                "If this continues to happen <a href='https://github.com/rlaphoenix/nton/issues'>Create an Issue</a>.",
                QMessageBox.StandardButton.Ok,
            )
            sys.exit(1)
    if not Files.game_title_ids.exists():
        QMessageBox.critical(
            window,
            "Corrupt Installation",
            "Could not Game Title ID registry which should be pre-bundled with NTON.<br/><br/>"
            "If this continues to happen <a href='https://github.com/rlaphoenix/nton/issues'>Create an Issue</a>.",
            QMessageBox.StandardButton.Ok,
        )
        sys.exit(1)
    while not Files.keys.is_file():
        QMessageBox.critical(
            window,
            "No prod.keys",
            f'Could not find prod.keys file. Place it in the installation directory, or at "{Files.keys_home}".'
            "<br/><br/>"
            "Press OK once done to continue.",
            QMessageBox.StandardButton.Ok,
        )

    # TODO: Add GUI menu bar option to update Game Title ID registry, then warn if outdated
    # if Files.game_title_ids.stat().st_mtime + (60 * 24 * 30) < time.time():
    #     log.warning("Game Title ID registry is quite old, I recommend updating it with `nton update-game-ids`")


def setup_logic(window: MainWindow) -> None:
    # menu bar actions
    window.ui.actionOpen.triggered.connect(partial(open_nro_file, window))
    window.ui.actionClose.triggered.connect(partial(reset_ui, window))
    window.ui.actionExit.triggered.connect(window.close)
    window.ui.actionTroubleshooting.triggered.connect(
        lambda: webbrowser.open("https://github.com/rlaphoenix/nton#troubleshooting")
    )
    window.ui.actionAbout.triggered.connect(partial(about, window))

    # main ui fields
    window.ui.name.textChanged.connect(partial(update_control_nacp, 0x0, 0x200))
    window.ui.author.textChanged.connect(partial(update_control_nacp, 0x200, 0x100))
    window.ui.displayVersion.textChanged.connect(partial(update_control_nacp, 0x3060, 0x10))
    window.ui.videoCapture.currentIndexChanged.connect(partial(update_control_nacp, 0x3035, 0x01))
    window.ui.screenshots.currentIndexChanged.connect(partial(update_control_nacp, 0x3034, 0x01))
    window.ui.titleId.textChanged.connect(partial(title_id_validator, window))

    # main ui buttons
    window.ui.randomizeIdButton.clicked.connect(lambda: window.ui.titleId.setText(generate_title_id()))
    window.ui.loadRomButton.clicked.connect(partial(set_rom_args, window))
    window.ui.buildButton.clicked.connect(partial(build, window))

    # drag/drop to load nro
    def setOpenTextColor(color: str) -> None:
        window.ui.openLabel.setStyleSheet(f"border: 3px solid {color};\nborder-style: dashed;\nborder-radius: 8px;")

    nroDropper = FileDropper(window)
    nroDropper.onHover.connect(partial(setOpenTextColor, "#0868c0"))
    nroDropper.onLeave.connect(partial(setOpenTextColor, "#dcdcdc"))
    nroDropper.dropped.connect(partial(load_nro_file, window))
    nroDropper.clicked.connect(partial(open_nro_file, window))
    window.ui.appPage.setAcceptDrops(True)
    window.ui.appPage.installEventFilter(nroDropper)

    # drag/drop to load image
    imageDropper = FileDropper(window)
    imageDropper.dropped.connect(partial(load_icon_file, window))
    imageDropper.clicked.connect(partial(open_icon_file, window))
    window.ui.icon.setAcceptDrops(True)
    window.ui.icon.installEventFilter(imageDropper)


def reset_ui(window: MainWindow) -> None:
    """Reset the UI to initial startup state."""
    global NRO_PATH
    global SDMC
    global CONTROL_NACP

    NRO_PATH = None
    SDMC = None
    CONTROL_NACP = bytearray()

    window.ui.appPage.setCurrentIndex(0)

    window.ui.openLabel.setStyleSheet("border: 3px solid #dcdcdc;\nborder-style: dashed;\nborder-radius: 8px;")

    window.ui.name.setText("")
    window.ui.author.setText("")
    window.ui.icon.setPixmap(QPixmap())
    window.ui.version.setValue(0)
    window.ui.displayVersion.setText("")
    window.ui.videoCapture.setCurrentIndex(0)
    window.ui.screenshots.setCurrentIndex(0)
    window.ui.titleId.setText("")

    disable_all_fields(window)

    window.ui.statusbar.showMessage("Ready")


def disable_all_fields(window: MainWindow) -> None:
    """Disable all modifiable fields."""
    window.ui.name.setEnabled(False)
    window.ui.author.setEnabled(False)
    window.ui.icon.setEnabled(False)
    window.ui.version.setEnabled(False)
    window.ui.displayVersion.setEnabled(False)
    window.ui.videoCapture.setEnabled(False)
    window.ui.screenshots.setEnabled(False)
    window.ui.titleId.setEnabled(False)


def enable_all_fields(window: MainWindow) -> None:
    """Enable all modifiable fields."""
    window.ui.name.setEnabled(True)
    window.ui.author.setEnabled(True)
    window.ui.icon.setEnabled(True)
    # https://github.com/The-4n/hacBrewPack/issues/15
    # window.version.setEnabled(True)
    window.ui.displayVersion.setEnabled(True)
    window.ui.videoCapture.setEnabled(True)
    window.ui.screenshots.setEnabled(True)
    window.ui.titleId.setEnabled(True)


def about(window: MainWindow) -> None:
    """Displays the Help->About Message Box."""
    QMessageBox.about(
        window,
        "About",
        f"<h1><strong>NTON v{__version__}</strong></h1>"
        + "<p>"
        + f"Version {__version__}<br/>"
        + f"Copyright &copy; {get_copyright_years()} &nbsp; rlaphoenix<br/>"
        + "GNU General Public License, Version 3.0."
        + "</p>"
        + "<a href='https://github.com/rlaphoenix/nton' style='color:blue'>https://github.com/rlaphoenix/nton</a>"
        + "<br/>"
        + "<hr/>"
        + "<p>"
        + " • hacBrewPack licensed under GPL-v2 for packing the NSP:<br/>"
        + "&nbsp;&nbsp; <a href='https://github.com/The-4n/hacBrewPack' style='color:blue'>https://github.com/The-4n/hacBrewPack</a><br/>"
        + " • nstool licensed under MIT for NRO extraction and verification:<br/>"
        + "&nbsp;&nbsp; <a href='https://github.com/jakcron/nstool' style='color:blue'>https://github.com/jakcron/nstool</a><br/>"
        + " • hptnacp from hacPack licensed under GPL-v2 for creating new NACP partitions if the NRO did not have one:<br/>"
        + "&nbsp;&nbsp; <a href='https://github.com/The-4n/hacPack/tree/master/hacPack-Tools/hacPackTools-NACP' style='color:blue'>https://github.com/The-4n/hacPack/tree/master/hacPack-Tools/hacPackTools-NACP</a><br/>"
        + "<br/>"
        + "No changes were made to any of the aforementioned software.<br/>"
        + "<br/>"
        + ' • <a href="https://flaticon.com/free-icons/open-folder">Open folder icons created by Freepik - Flaticon</a><br/>'
        + ' • <a href="https://flaticon.com/free-icons/close">Close icons created by Freepik - Flaticon</a><br/>'
        + ' • <a href="https://flaticon.com/free-icons/delete">Delete icons created by Pixel perfect - Flaticon</a>'
        + "</p>"
        + "<hr/>"
        + "<br/>"
        + 'Made with ❤️ in Ireland &nbsp; <span style="color:green">█</span><span style="color:white">█</p><span style="color:orange">█</span>',
    )


def open_nro_file(window: MainWindow) -> bool:
    """Open a File Dialogue to open and load a file as an NRO."""
    loc = QFileDialog.getOpenFileName(window, "Homebrew Applet (NRO)", "", "NRO files (*.nro);;All files (*.*)")
    if not loc[0]:
        return False

    file_path = Path(loc[0])
    log.debug("Opened File %s", file_path)

    return load_nro_file(window, file_path)


def load_nro_file(window: MainWindow, file_path: Path) -> bool:
    """
    Load NRO file by path.

    This calculates the SDMC path (path to NRO relative to microSD card), loads the data
    to the UI fields, and sets the Stacked Widget to page 2 (the main file-loaded UI).
    """
    global SDMC
    global NRO_PATH

    reset_ui(window)

    window.ui.statusbar.showMessage(f"Loading {file_path.name}")
    log.info("Loading NRO: %s", file_path)
    if not load_nro_data(window, file_path):
        window.ui.statusbar.showMessage(f"Failed loading {file_path.name}")
        log.error("Failed loading NRO: %s", file_path)
        return False

    NRO_PATH = file_path

    path_root = Path(file_path.anchor)
    if not (
        ((path_root / "Nintendo").exists() and (path_root / "switch").exists())
        or ((path_root / "atmosphere").exists() and ((path_root / "bootloader").exists()))
    ):
        log.debug("Couldn't find a Switch-related folder in %s", path_root)
        sdmc, ok = QInputDialog.getText(
            window,
            "SDMC Path",
            "This NRO file is on your PC but NTON is intended to be used with NROs on your Switch's microSD card.<br/>"
            "Please specify where the NRO will be located on your Switch's microSD card. (e.g., /switch/daybreak.nro).",
        )
        sdmc = sdmc.strip()
        if not ok or not sdmc:
            log.debug("User did not give an SDMC")
            window.ui.statusbar.showMessage(f"Cancelled loading {file_path.name}")
            window.ui.appPage.setCurrentIndex(0)
            return False
        log.debug("SDMC given: %s", sdmc)
        if not sdmc.startswith("/"):
            sdmc = f"/{sdmc}"
        SDMC = Path(sdmc).resolve().absolute().as_posix().replace(file_path.anchor.replace("\\", "/"), "sdmc:/")
    else:
        SDMC = file_path.resolve().absolute().as_posix().replace(file_path.anchor.replace("\\", "/"), "sdmc:/")
    log.debug("Resolved SDMC: %s", SDMC)

    window.ui.appPage.setCurrentIndex(1)
    window.ui.statusbar.showMessage(f"Loaded {file_path.name}")
    log.info("Loaded NRO: %s", file_path)

    return True


def load_nro_data(window: MainWindow, nro_path: Path) -> bool:
    """Load NRO Name, Author, Icon, Version, and more to the UI."""
    global CONTROL_NACP

    if nro_path.stem.lower() in title_ids.unofficial:
        title_id = title_ids.unofficial[nro_path.stem.lower()]
        log.info("Got Title ID from NTON registry: %s", title_id)
    else:
        title_id = generate_title_id()
        log.info("Got Randomized Title ID: %s", title_id)
    window.ui.titleId.setText(title_id.upper())

    with tempfile.TemporaryDirectory(prefix="rlaphoenix-nton") as t:
        log.debug("Opened Temporary Directory: %s", t)
        load_dir = Path(t)

        control_file = load_dir / "control.nacp"
        icon_file = load_dir / "icon_AmericanEnglish.dat"

        control_file_res = nstool.get_nacp(nro_path, control_file)
        if control_file_res:
            log.info("Error loading NACP: %s", control_file_res)
            if control_file_res == "No NACP was extracted from the asset.":
                window.ui.name.setText(nro_path.stem)
                window.ui.icon.setPixmap(QPixmap(":/branding/images/sad.png"))
                window.ui.displayVersion.setText("v1.0.0")
                window.ui.videoCapture.setCurrentIndex(2)
                enable_all_fields(window)
                QMessageBox.information(
                    window,
                    "Notice: Crappy NRO",
                    f"The NRO file '{nro_path.name}' is poorly made; it has no NACP partition.<br/>"
                    "This means NTON cannot infer ANY information. Some common defaults have been set, and "
                    "some have been assumed based on the filename.<br/><br/>"
                    "Please note that this is not an error, but there are few NROs without a NACP "
                    "partition. If this is not expected, please verify the integrity of your NRO file.",
                    QMessageBox.StandardButton.Ok,
                )
                return True

            window.ui.appPage.setCurrentIndex(0)
            if "Input file type was undetermined" in str(control_file_res):
                QMessageBox.critical(
                    window,
                    "Failed to load file",
                    f"The file '{nro_path.name}' does not seem to be a valid NRO file,<br/><br/>{control_file_res}",
                    QMessageBox.StandardButton.Ok,
                )
            else:
                QMessageBox.critical(
                    window,
                    "Failed to load file",
                    f"The NRO file '{nro_path.name}' failed to load while extracting the NACP partition,<br/><br/>"
                    f"{control_file_res}<br/><br/>"
                    f"Cannot continue...",
                    QMessageBox.StandardButton.Ok,
                )
            return False

        CONTROL_NACP = bytearray(control_file.read_bytes())
        log.info("Extracted the Control NACP Partition")

        window.ui.videoCapture.setCurrentIndex(CONTROL_NACP[0x3035])
        log.info("Video Capture: %d", CONTROL_NACP[0x3035])
        window.ui.screenshots.setCurrentIndex(CONTROL_NACP[0x3034])
        log.info("Screenshots: %s", CONTROL_NACP[0x3034])

        language_data = {
            lang: {
                "name": CONTROL_NACP[offset : offset + 0x10].replace(b"\x00", b"").strip().decode("utf8"),
                "publisher": CONTROL_NACP[offset + 0x200 : offset + 0x210].replace(b"\x00", b"").strip().decode("utf8"),
            }
            for i, lang in enumerate(
                (
                    "AmericanEnglish",
                    "BritishEnglish",
                    "Japanese",
                    "French",
                    "German",
                    "LatinAmericanSpanish",
                    "Spanish",
                    "Italian",
                    "Dutch",
                    "CanadianFrench",
                    "Portuguese",
                    "Russian",
                    "Korean",
                    "TraditionalChinese",
                    "SimplifiedChinese",
                )
            )
            for offset in [0x0300 * i]
        }

        log.info("Language Data:")

        for lang, data in language_data.items():
            if not data["name"] and not data["publisher"]:
                continue
            log.info(' - %s: [Name: "%s", Publisher: "%s"]', lang, data["name"], data["publisher"])

        for lang, data in language_data.items():
            if data["name"]:
                # TODO: What if another is preferred?
                window.ui.name.setText(data["name"])
                window.ui.author.setText(data["publisher"])
                log.info("Chosen NRO name and publisher from %s Language Data", lang)
                break

        # only keep one name/publisher, store as AmericanEnglish
        # this is because the GUI has only one name/publisher field to use
        update_control_nacp(0x0, 0x300 * 0x10, 0x0)
        update_control_nacp(0x0, 0x200, window.ui.name.text())
        update_control_nacp(0x200, 0x100, window.ui.author.text())
        log.info("Removed all languages, set chosen data as AmericanEnglish")

        icon_file_res = nstool.get_icon(nro_path, icon_file)
        if icon_file_res:
            if icon_file_res == "No Icon was extracted from the asset.":
                log.info("The NRO did not have an Icon")
            else:
                log.error("Failed extracting the Icon partition from the NRO, %s", icon_file_res)
                res = QMessageBox.question(
                    window,
                    "Failed to extract Icon",
                    f"The NRO file '{nro_path.name}' failed to load while extracting the Icon,<br/><br/>"
                    f"{icon_file_res}",
                    QMessageBox.StandardButton.Cancel,
                    QMessageBox.StandardButton.Ignore,
                )
                if res != QMessageBox.StandardButton.Ignore:
                    sys.exit(2)
        else:
            log.info("Got the Icon from the NRO")
        window.ui.icon.setPixmap(QPixmap(icon_file))

        window.ui.displayVersion.setText(CONTROL_NACP[0x3060:0x306F].replace(b"\x00", b"").strip().decode("utf8"))
        log.info("Display Version: %s", window.ui.displayVersion.text())

        enable_all_fields(window)

        return True


def open_icon_file(window: MainWindow) -> None:
    loc = QFileDialog.getOpenFileName(
        window,
        "Homebrew Icon",
        "",
        "Image files (*.png *.jpg *.jpeg *.bmp *.tiff *.tif *.webp *.heif);;All files (*.*)",
    )
    if loc[0]:
        file_path = Path(loc[0])
        load_icon_file(window, file_path)


def load_icon_file(window: MainWindow, icon_path: Path) -> None:
    window.ui.statusbar.showMessage(f"Loading Icon {icon_path.name}")
    window.ui.icon.setPixmap(QPixmap(icon_path))
    window.ui.statusbar.showMessage(f"Set {icon_path.name} as Icon")


def title_id_validator(window: MainWindow, value: str) -> None:
    """Ensure the Title ID is Valid."""
    if RE_TITLE_ID.match(value):
        window.ui.titleId.setStyleSheet("")
    else:
        window.ui.titleId.setStyleSheet("color:red")


def generate_title_id() -> str:
    """
    Generate a Random Title ID.

    The Title ID is ensured to begin with `01` and end with `000`.
    A Base Title typically ends with `000` while a Patch ends with `800`.
    """
    title_id = "0100000000000000"
    while title_id in title_ids.ALL:
        title_id = "01%s000" % os.urandom(6).hex()[:-1]
    return title_id


def update_control_nacp(offset: int, length: int, value: Union[int, str, bytes]) -> None:
    """
    Overwrite Data on the Control NACP at specific offsets.

    If the specified length is larger than the value itself, then it will be
    right-padded with 0x00. If the value is a string, it will be UTF-8 encoded.
    If the value is a number, it will be encoded to bytes in little-endian order.
    """
    if isinstance(value, str):
        vb = value.encode("utf8")
    elif isinstance(value, int):
        vb = value.to_bytes(length, "little")
    elif not isinstance(value, bytes):
        raise TypeError("Value must be int, str, or bytes")
    if len(vb) < length:
        vb = vb.ljust(length, b"\x00")
    CONTROL_NACP[offset : offset + length] = vb


def set_rom_args(window: MainWindow) -> None:
    """
    Helper utility for writing a ROM argv to load a SDMC file.

    This function is not actually specific for Game ROMs, it can be used as a helper
    to write the args to load any kind of file. It just specifies Game ROM just for
    UX reasons as Forwarders mainly use arguments for Game ROM specification.
    """
    loc = QFileDialog.getOpenFileName(
        window,
        "Game ROM (from SD card!)",
        "",
        "Game ROMs (*.3ds *.a26 *.a52 *.amiga *.amstrad *.atari *.c64 *.cpc *.dcm *.fds *.gamegear *.gb *.gbc *.gba"
        "*.gdi *.gg *.intellivision *.lynx *.mame *.md *.msx *.nds *.neogeo *.nes *.ngp *.ngpc *.pce *.pcfx *.pico"
        "*.ps1 *.ps2 *.ps3 *.psp *.psvita *.rom *.sfc *.sg1000 *.sgb *.sms *.snes *.virtualboy *.wii *.wiiu *.xbox);;"
        "All Files (*)",
    )
    if not loc[0]:
        return

    rom_path = Path(loc[0])

    path_root = Path(rom_path.anchor)
    if not (
        ((path_root / "Nintendo").exists() and (path_root / "switch").exists())
        or ((path_root / "atmosphere").exists() and ((path_root / "bootloader").exists()))
    ):
        sdmc, ok = QInputDialog.getText(
            window,
            "ROM Path",
            "This ROM file is on your PC but NTON is intended to be used with ROMs on your Switch's microSD card.<br/>"
            "Please specify where the ROM will be located on your Switch's microSD card. (e.g., /roms/system/game.gb).",
        )
        sdmc = sdmc.strip()
        if not ok or not sdmc:
            return
        if not sdmc.startswith("/"):
            sdmc = f"/{sdmc}"
        sdmc = Path(sdmc).resolve().absolute().as_posix().replace(rom_path.anchor.replace("\\", "/"), "sdmc:/")
    else:
        sdmc = rom_path.resolve().absolute().as_posix().replace(rom_path.anchor.replace("\\", "/"), "sdmc:/")

    existing_args = window.ui.args.text().strip()
    if existing_args:
        existing_args += " "
    existing_args += f'"{sdmc}"'
    window.ui.args.setText(existing_args)


def build(window: MainWindow) -> bool:
    """Build the NSP using the currently loaded Control NACP and UI fields."""
    if not Binaries.hacbrewpack:
        QMessageBox.critical(
            window,
            "Missing Dependency",
            "The hacBrewPack binary is missing, which is required to build NSP files.<br/><br/>"
            "Please reinstall NTON or visit <a href='https://github.com/switchbrew/hacBrewPack'>https://github.com/switchbrew/hacBrewPack</a>",
            QMessageBox.StandardButton.Ok,
        )
        return False
    if not NRO_PATH:
        QMessageBox.critical(
            window,
            "No NRO Selected",
            "No NRO file is currently selected. Please load an NRO file before building an NSP.",
        )
        return False
    if not SDMC:
        QMessageBox.critical(
            window,
            "No SDMC Path",
            "No SDMC path is currently set. Please load an NRO file and set its SDMC path before building an NSP.",
        )
        return False

    name = window.ui.name.text()
    author = window.ui.author.text()
    version = window.ui.version.text()
    title_id = window.ui.titleId.text().lower()

    if any(c not in string.hexdigits for c in title_id):
        QMessageBox.critical(
            window,
            "Title ID Error",
            f'The Title ID "{title_id}" is an invalid hex string. It must be a-fA-F0-9.<br/><br/>'
            "Build cannot continue...",
        )
        return False
    if not RE_TITLE_ID.match(title_id):
        QMessageBox.critical(
            window,
            "Title ID Error",
            f'The Title ID "{title_id}" is not a valid range or length. It must begin with 01 and end with 000.'
            "<br/><br/>Build cannot continue...",
        )
        return False
    if title_id in title_ids.ALL_SYSTEM:
        QMessageBox.critical(
            window,
            "System Title ID Conflict",
            f'The Title ID "{title_id}" is a reserved System Title! It cannot be used!<br/><br/>'
            "Build cannot continue...",
        )
        return False
    if title_id in title_ids.game_title_ids:
        res = QMessageBox.question(
            window,
            "Game Title ID Conflict",
            f'The Title ID "{title_id}" is already used by "{title_ids.game_title_ids[title_id]}".<br/><br/>'
            "Are you sure you want to continue?",
        )
        if res != QMessageBox.StandardButton.Yes:
            return False

    verification = nstool.verify(NRO_PATH, "nro")
    if verification:
        log.error("Integrity Error on the NRO file: %s", verification)
        QMessageBox.critical(
            window,
            "Integrity Error",
            f'The NRO "{NRO_PATH}" is invalid, {verification}<br/><br/>Build cannot continue...',
        )
        return False
    log.info("NRO Integrity: OK")

    save_filename = QFileDialog.getSaveFileName(
        window,
        "Build NSP",
        str(Directories.output / f"{name} by {author} [{title_id.upper()}][v{version}].nsp"),
        "NSP (*.nsp);;All files (*.*)",
    )[0]
    if not save_filename:
        log.info("Cancelled Build as no file save path was chosen")
        return False

    save_path = Path(save_filename)
    log.debug("Chosen Save Path: %s", save_path)

    build_dir = Directories.temp / title_id
    control_dir = build_dir / "control"
    romfs_dir = build_dir / "romfs"
    exefs_dir = build_dir / "exefs"
    logo_dir = build_dir / "logo"
    hacbrewpack_backup_dir = build_dir / "hacbrewpack_backup"
    control_file = control_dir / "control.nacp"
    icon_file = control_dir / "icon_AmericanEnglish.dat"
    next_argv_file = romfs_dir / "nextArgv"
    next_nro_path_file = romfs_dir / "nextNroPath"

    try:
        if build_dir.exists():
            log.debug("Build Directory already exists, somehow, deleting...")
            shutil.rmtree(build_dir)
        build_dir.mkdir(parents=True)
        control_dir.mkdir()
        romfs_dir.mkdir()

        shutil.copytree(Directories.assets / "exefs", exefs_dir)
        shutil.copytree(Directories.assets / "logo", logo_dir)

        # disable user profile selection as it's unnecessary
        update_control_nacp(0x3025, 0x01, 0x0)

        # disable all storage allocation to ensure a 1MB install size
        update_control_nacp(0x3080, 0x08, 0x00)  # User Account Save Data
        update_control_nacp(0x3088, 0x08, 0x00)  # User Account Save Data Journal
        update_control_nacp(0x3090, 0x08, 0x00)  # Device Save Data
        update_control_nacp(0x3098, 0x08, 0x00)  # Device Save Data Journal
        update_control_nacp(0x30A0, 0x08, 0x00)  # BCAT Delivery Cache Storage
        update_control_nacp(0x3148, 0x08, 0x00)  # Max User Account Save Data
        update_control_nacp(0x3150, 0x08, 0x00)  # Max User Account Save Data Journal
        update_control_nacp(0x3158, 0x08, 0x00)  # Max Device Save Data
        update_control_nacp(0x3160, 0x08, 0x00)  # Max Device Save Data Journal
        update_control_nacp(0x3168, 0x08, 0x00)  # Temporary Storage
        update_control_nacp(0x3170, 0x08, 0x00)  # Cache Storage
        update_control_nacp(0x3178, 0x08, 0x00)  # Cache Storage Journal
        update_control_nacp(0x3180, 0x08, 0x00)  # Max Cache Storage Data and Journal
        update_control_nacp(0x3188, 0x02, 0x00)  # Max Cache Storage Index

        control_file.write_bytes(CONTROL_NACP)
        log.debug("Written %d bytes to Control NACP file", len(CONTROL_NACP))
        window.ui.icon.pixmap().save(str(icon_file), "JPG")

        if icon_file.exists():
            # We must strip every unnecessary metadata or the icon will be a '?'
            im = Image.open(icon_file)
            if im.size != (256, 256):
                im = im.resize((256, 256))
            if im.mode != "RGB":
                im = im.convert("RGB")
            clean_im = Image.new(im.mode, im.size)
            clean_im.putdata(list(im.getdata()))
            clean_im.save(icon_file, format="JPEG")
            clean_im.close()
            im.close()
            log.debug("Converted and Stripped Icon File")

        next_nro_path_file.write_text(SDMC)

        next_argv = SDMC
        args = window.ui.args.text().strip()
        if args:
            next_argv += f" {args}"
        next_argv_file.write_text(next_argv)

        # only make this directory at this point because we have a high chance of success
        Directories.output.mkdir(parents=True, exist_ok=True)

        try:
            startupinfo = subprocess.STARTUPINFO()
            startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
            subprocess.check_output(
                [
                    Binaries.hacbrewpack,
                    "--titleid",
                    title_id,
                    "--nspdir",
                    str(Directories.output.absolute()),
                    "-k",
                    str(Files.keys.absolute()),
                ],
                cwd=build_dir,
                startupinfo=startupinfo,
            )
        except subprocess.CalledProcessError as e:
            log.error('Failed to build NSP, "%s", %s [%d]', e.args, e.output, e.returncode)
            QMessageBox.critical(
                window, "Build Failed", f'Failed to build NSP, "{e.args}", {e.output} [{e.returncode}]'
            )
            return False

        if save_path.exists():
            save_path.unlink()
        shutil.move(Directories.output / f"{title_id}.nsp", save_path)

        log.info("Build NSP to %s", save_path)
        QMessageBox.information(window, "Success", f'An NSP forwarder was built to "{save_path}".')
        return True
    finally:
        if build_dir.exists():
            shutil.rmtree(build_dir)
        if hacbrewpack_backup_dir.exists():
            shutil.rmtree(hacbrewpack_backup_dir)


if __name__ == "__main__":
    start()



import os
import shutil
import tempfile
from pathlib import Path


class Directories:
    root = Path(__file__).resolve().parent  # root of package/src
    temp = Path(tempfile.gettempdir()) / "rlaphoenix-nton"
    output = Path.home() / "Desktop" / "NTON"
    assets = root / "assets"
    bin = root / "bin"


PATH = os.environ.get("PATH", "")
PATH = os.pathsep.join([str(Directories.bin), PATH])
os.environ["PATH"] = PATH


class Binaries:
    nstool = shutil.which("nstool")
    hacbrewpack = shutil.which("hacbrewpack")
    hptnacp = shutil.which("hptnacp")


class Files:
    keys_home = Path.home() / ".switch" / "prod.keys"
    keys_cwd = Path("./prod.keys").absolute()
    keys = keys_cwd if keys_cwd.is_file() else keys_home
    game_title_ids = Directories.assets / "game_title_ids.json"
    ui_file = Directories.root / "gui" / "main.ui"



from pathlib import Path
from typing import Any, cast

from PySide6.QtCore import QEvent, QObject, Signal
from PySide6.QtGui import QDragEnterEvent, QDropEvent


class FileDropper(QObject):
    dropped = Signal(Path)  # has file dropped
    onHover = Signal()  # is hovered
    onLeave = Signal()  # is no longer hovered
    clicked = Signal()  # is clicked

    def eventFilter(self, obj: Any, event: QEvent) -> bool:
        if event.type() == QEvent.Type.DragEnter:
            dragEvent: QDragEnterEvent = cast(QDragEnterEvent, event)
            if dragEvent.mimeData().hasUrls():
                dragEvent.accept()
                self.onHover.emit()
                return True

        elif event.type() == QEvent.Type.DragLeave:
            self.onLeave.emit()
            return True

        elif event.type() == QEvent.Type.Drop:
            dropEvent: QDropEvent = cast(QDropEvent, event)
            for url in dropEvent.mimeData().urls():
                path = url.toLocalFile()
                if path:
                    self.dropped.emit(Path(path))
            return True

        elif event.type() == QEvent.Type.MouseButtonPress:
            self.clicked.emit()
            return True

        return False



from datetime import datetime


def get_copyright_years() -> str:
    start_year = 2022
    current_year = datetime.now().year
    if start_year == current_year:
        return str(start_year)
    return f"{start_year}-{current_year}"



import itertools
import shutil
from argparse import ArgumentParser
from pathlib import Path
from textwrap import dedent

from PyInstaller.__main__ import run

from nton import __version__
from nton.helpers import get_copyright_years

parser = ArgumentParser()
parser.add_argument("--debug", action="store_true", help="Enable debug mode (keeps leftover build files)")
parser.add_argument("--name", default="NTON", help="Set the Project Name")
parser.add_argument("--author", default="rlaphoenix", help="Set the Project Author")
parser.add_argument("--version", default=__version__, help="Set the EXE Version")
parser.add_argument(
    "--icon-file", default="nton/gui/resources/images/nton.ico", help="Set the Icon file path (must be a .ICO file)"
)
parser.add_argument("--one-file", action="store_true", help="Build to a singular .exe file")
parser.add_argument("--console", action="store_true", help="Show the Console window")
args = parser.parse_args()


"""Configuration options that may be changed or referenced often."""
ADDITIONAL_DATA = [
    # local file path, destination in build output
    ["nton/assets", "nton/assets"],
    ["nton/bin", "nton/bin"],
    ["nton/gui/main.ui", "nton/gui"],
]
EXTRA_ARGS = ["-y"]

"""Prepare environment to ensure output data is fresh."""
shutil.rmtree("build", ignore_errors=True)
shutil.rmtree("dist", ignore_errors=True)
Path("NTON.spec").unlink(missing_ok=True)
version_file = Path("pyinstaller.version.txt")


"""Set Version Info Structure."""
VERSION_4_TUP = tuple(map(int, ("%s.0" % args.version).split(".")))
version_file.write_text(
    dedent(f"""
    VSVersionInfo(
      ffi=FixedFileInfo(
        filevers=({", ".join(args.version.split("."))}, 0),
        prodvers=({", ".join(args.version.split("."))}, 0),
        OS=0x40004,  # Windows NT
        fileType=0x1,  # Application
        subtype=0x0  # type is undefined
      ),
      kids=[
        StringFileInfo(
          [
          StringTable(
            '040904b0',
            [StringStruct('CompanyName', '{args.author}'),
            StringStruct('FileDescription', 'Nintendo Switch NRO to NSP Forwarder'),
            StringStruct('FileVersion', '{args.version}'),
            StringStruct('InternalName', '{args.name}'),
            StringStruct('LegalCopyright', '{f"Copyright (C) {get_copyright_years()} %s" % args.author}'),
            StringStruct('OriginalFilename', 'NTON.exe'),
            StringStruct('ProductName', '{args.name}'),
            StringStruct('ProductVersion', '{args.version}'),
            StringStruct('Comments', '{args.name}')])
          ]),
        VarFileInfo([VarStruct('Translation', [1033, 1200])])
      ]
    )
    """).strip(),
    encoding="utf8",
)

"""Run PyInstaller with the provided configuration."""
try:
    run(
        [
            "nton/__main__.py",
            "-n",
            args.name,
            "-i",
            ["NONE", args.icon_file][bool(args.icon_file)],
            ["-D", "-F"][args.one_file],
            ["-w", "-c"][args.console],
            *itertools.chain(*[["--add-data", ":".join(x)] for x in ADDITIONAL_DATA]),
            "--version-file",
            str(version_file),
            *EXTRA_ARGS,
        ]
    )
finally:
    if not args.debug:
        shutil.rmtree("build", ignore_errors=True)
        Path("NTON.spec").unlink(missing_ok=True)
        version_file.unlink(missing_ok=True)



from __future__ import annotations

import base64
import logging
import os
import shutil
import string
import subprocess
import sys
import time
from pathlib import Path

import click as click
import coloredlogs
import jsonpickle
from bs4 import BeautifulSoup
from PIL import Image

from nton import __version__, nstool, title_ids
from nton.constants import Binaries, Directories, Files
from nton.helpers import get_copyright_years
from nton.title_ids import get_game_title_ids


@click.group(invoke_without_command=True)
@click.option("-v", "--version", is_flag=True, default=False, help="Print version information.")
@click.option("-d", "--debug", is_flag=True, default=False, help="Enable DEBUG level logs.")
@click.pass_context
def main(ctx: click.Context, version: bool, debug: bool) -> None:
    """nton—Nintendo Switch NRO to NSP Forwarder."""
    logging.basicConfig(level=logging.DEBUG if debug else logging.INFO)
    log = logging.getLogger(__name__)
    coloredlogs.install(
        level=(log.parent or log).level, logger=log, fmt="{asctime} [{levelname[0]}] {name} : {message}", style="{"
    )

    log.info("nton version %s Copyright (c) %s rlaphoenix", __version__, get_copyright_years())
    log.info("https://github.com/rlaphoenix/nton")
    if version:
        return

    if not ctx.invoked_subcommand:
        from nton import gui  # noqa

        gui.start(debug)
    else:
        for binary, path in vars(Binaries).items():
            if binary.startswith("__"):
                continue
            if not path:
                log.error(
                    f"{binary} cannot be found! Place it in the current working directory, "
                    "or put the folder it's at into your PATH Environment Variable."
                )
                sys.exit(1)
        if not Files.keys.is_file():
            log.error(
                f'prod.keys cannot be found! Place it in the current working directory, or at "{Files.keys_home}"'
            )
            sys.exit(1)
        if not Files.game_title_ids.exists():
            log.error("Game Title ID registry is missing! Please re-add `/assets/game_title_ids.json`!")
            sys.exit(1)
        if Files.game_title_ids.stat().st_mtime + (60 * 24 * 30) < time.time():
            log.warning("Game Title ID registry is quite old, I recommend updating it with `nton update-game-ids`")


@main.command()
@click.argument("path", type=Path)
@click.option("-n", "--name", type=str, default=None, help="Title Name.")
@click.option("-p", "--publisher", type=str, default=None, help="Title Publisher.")
@click.option("-v", "--version", type=str, default=None, help="Title Version.")
@click.option("-i", "--icon", type=Path, default=None, help="Title Icon (256x256px recommended, supports any image).")
@click.option("--id", "id_", type=str, default=None, help="Title ID.")
@click.option("--rom", type=str, default=None, help="ROM path for Direct Game Forwarding.")
@click.option("--sdmc", type=str, default=None, help="NRO path relative to the root of the Switch's microSD card.")
def build(
    path: Path,
    name: str | None,
    publisher: str | None,
    version: str | None,
    icon: Path | None,
    id_: str | None,
    rom: str | None,
    sdmc: str | None,
) -> None:
    """
    Build an NSP that loads an NRO on the Switch's microSD card.

    \b
    Parameters:
        path: Path to the NRO file. You MUST provide the path on the Switch's microSD card, NOT a file on your PCs
            HDD! This is so we can automatically get the sdmc:/ path in a foolproof way.
        name: Override the Title name that is automatically retrieved from the NRO.
        publisher: Override the Publisher name that is automatically retrieved from the NRO.
        icon: Override the Icon that is automatically retrieved from the NRO. This can be any image of any format or
            resolution. A 256x256px image is recommended.
        id_: Set a specific Title ID, otherwise a Random Title ID is used. There's a miniscule chance it could get the
            same Title ID as another installed Title, but it's so miniscule you shouldn't realistically worry about it.
        rom: Path to a ROM file on the Switch's microSD card to create a forwarder that boots directly into the game.
            The path to the ROM must be an absolute path. The NRO used must be a homebrew that supports specifying a
            ROM by args (e.g., a RetroArch Core, MGBA, possibly others).
        sdmc: Path to the NRO path relative to the root of the Switch's microSD card. This should only be used if the
            NRO path you provided is NOT on the microSD card, as it is implicitly inferred.
    """
    log = logging.getLogger("build")
    log.info("Building!")

    if not Binaries.hptnacp:
        log.error("hptnacp binary is missing, cannot build NSP.")
        sys.exit(1)

    if not Binaries.hacbrewpack:
        log.error("hacBrewPack binary is missing, cannot build NSP.")
        sys.exit(1)

    if not path.is_file():
        log.error(f'The NRO path "{path}" does not exist, or is not a file.')
        sys.exit(1)

    if path.suffix.lower() != ".nro":
        log.error(f'The NRO path "{path}" is not to an NRO file.')
        sys.exit(1)

    if sdmc:
        if not sdmc.startswith("sdmc:/"):
            if not sdmc.startswith("/"):
                sdmc = f"/{sdmc}"
            sdmc = f"sdmc:{sdmc}"
    else:
        path_root = Path(path.anchor)
        if not (
            ((path_root / "Nintendo").exists() and (path_root / "switch").exists())
            or ((path_root / "atmosphere").exists() and ((path_root / "bootloader").exists()))
        ):
            log.error("The NRO path must be a path on your Switch's microSD card to implicitly infer the sdmc path.")
            log.error("You can use --sdmc to manually specify the path relative to the Switch's microSD card.")
            sys.exit(1)
        sdmc = path.resolve().absolute().as_posix().replace(path.anchor.replace("\\", "/"), "sdmc:/")

    verification = nstool.verify(path, "nro")
    if verification:
        log.critical('The NRO "%s" is invalid, %s', path, verification)
        sys.exit(2)

    log.info("NRO checked and verified")

    if id_:
        if any(c not in string.hexdigits for c in id_):
            log.error(f'The Title ID "{id_}" is an invalid hex string. It must be a-fA-f0-9.')
            sys.exit(1)
        id_ = id_.lower()
        if id_ in title_ids.ALL_SYSTEM:
            log.critical(f'The Title ID "{id_}" is a reserved System Title! Using it is unsafe!')
            sys.exit(2)
        if id_ in title_ids.game_title_ids:
            log.warning(f'The Title ID "{id_}" is already used by "{title_ids.game_title_ids[id_]}".')
    elif path.stem.lower() in title_ids.unofficial:
        id_ = title_ids.unofficial[path.stem.lower()]
    elif sdmc and Path(sdmc).stem.lower() in title_ids.unofficial:
        id_ = title_ids.unofficial[Path(sdmc).stem.lower()]
    else:
        id_ = "0100000000000000"
        while id_ in title_ids.ALL:
            id_ = "01%s000" % os.urandom(6).hex()[:-1]

    if rom and not rom.startswith("/"):
        rom = f"/{rom}"

    log.info("Title ID: %s", id_)

    build_dir = Directories.temp / id_
    control_template_file = Directories.assets / "control.nacp.xml"
    control_dir = build_dir / "control"
    romfs_dir = build_dir / "romfs"
    exefs_dir = build_dir / "exefs"
    logo_dir = build_dir / "logo"
    hacbrewpack_backup_dir = build_dir / "hacbrewpack_backup"
    control_file = control_dir / "control.nacp"
    icon_file = control_dir / "icon_AmericanEnglish.dat"
    next_argv_file = romfs_dir / "nextArgv"
    next_nro_path_file = romfs_dir / "nextNroPath"

    log.debug("Build Directory: %s", build_dir)

    try:
        if build_dir.exists():
            log.debug("Build Directory pre-existing")
            shutil.rmtree(build_dir)
        build_dir.mkdir(parents=True)
        control_dir.mkdir()
        romfs_dir.mkdir()

        shutil.copytree(Directories.assets / "exefs", exefs_dir)
        shutil.copytree(Directories.assets / "logo", logo_dir)

        control_file_res = nstool.get_nacp(path, control_file)
        if control_file_res:
            if control_file_res == "No NACP was extracted from the asset.":
                log.warning("The NRO does not have a NACP partition, building a new one.")
                if not name or not publisher or not version:
                    log.error("You must specify a Name, Publisher, and Version to be able to build the NACP.")
                    log.error("You may also want to specify an Icon but it is not strictly necessary.")
                    sys.exit(1)
                control_template = control_template_file.read_text(encoding="utf8")
                root = BeautifulSoup(control_template, "xml")
                for title in root.find_all("Title"):
                    if title.Name:
                        title.Name.string = name
                    if title.Publisher:
                        title.Publisher.string = publisher
                for title_id in root.find_all(["PresenceGroupId", "SaveDataOwnerId", "LocalCommunicationId"]):
                    title_id.string = f"0x{id_}"
                displayVersion = root.find("DisplayVersion")
                if displayVersion:
                    displayVersion.string = version
                addOnContentBaseId = root.find("AddOnContentBaseId")
                if addOnContentBaseId:
                    addOnContentBaseId.string = hex(int(id_, 16) + 0x1000)
                tmp_control_template = build_dir / "control.nacp.xml"
                tmp_control_template.write_text(str(root))
                try:
                    subprocess.check_output(
                        [
                            Binaries.hptnacp,
                            "-a",
                            "createnacp",
                            "-i",
                            str(tmp_control_template.absolute()),
                            "-o",
                            str(control_file.absolute()),
                        ]
                    )
                except subprocess.CalledProcessError as e:
                    log.critical(f"Failed to build a new NACP, {e.output} [{e.returncode}]")
                    sys.exit(2)
                log.info("Built a new NACP")
            else:
                log.critical(f"Failed extracting the NACP partition from the NRO, {control_file_res}")
                sys.exit(2)

        control_file_data = bytearray(control_file.read_bytes())
        log.debug("Got the Control partition")
        log.debug(base64.b64encode(control_file_data).decode())

        # enable video capture and screenshots if disabled
        if control_file_data[0x3035] != 0x02:
            # 0x00 = Disabled, 0x01 = Manual, 0x02 = Enabled
            control_file_data[0x3035] = 0x02
            log.info("Enabled Video Capture")
        if control_file_data[0x3034] != 0x00:
            # 0x00 = Enabled, 0x01 = Disabled
            control_file_data[0x3034] = 0x00
            log.info("Enabled Screenshots")

        # disable storage allocation as none of them would be used
        save_data_size_offsets = {
            0x3080: "User Account Save Data",
            0x3088: "User Account Save Data Journal",
            0x3090: "Device Save Data",
            0x3098: "Device Save Data Journal",
            0x30A0: "BCAT Delivery Cache Storage",
            0x3148: "Max User Account Save Data",
            0x3150: "Max User Account Save Data Journal",
            0x3158: "Max Device Save Data",
            0x3160: "Max Device Save Data Journal",
            0x3168: "Temporary Storage",
            0x3170: "Cache Storage",
            0x3178: "Cache Storage Journal",
            0x3180: "Max Cache Storage Data and Journal",
        }
        for offset, offset_name in save_data_size_offsets.items():
            save_data_size = int.from_bytes(control_file_data[offset : offset + 8], byteorder="little")
            if save_data_size != 0:
                control_file_data[offset : offset + 8] = b"\x00" * 8
                log.info(f"Removed {offset_name} Allocation")

        # set cache storage index max to 0
        control_file_data[0x3188 : 0x3188 + 0x2] = b"\x00\x00"

        # disable user profile selection as it's unnecessary
        control_file_data[0x3025] = 0x00

        if not name:
            # TODO: Assumes "AmericanEnglish" name is the one that's used and wanted
            name = control_file_data[0x0000:0x000F].replace(b"\x00", b"").strip().decode("utf8")
        if not name:
            log.error("The Control Partition does not have any listed Name nor was one manually specified.")
            sys.exit(1)
        if not any(x in string.ascii_letters + string.digits for x in name):
            log.error(f'The Application Name, "{name}", cannot be all special characters.')
            sys.exit(1)
        if len(name.encode("utf8")) > 0x200:  # fits 0x200 * 10 (16 fields)
            log.error(f'The Title Name "{name}" is too large to fit in the NSP.')
            sys.exit(1)
        log.info("Title Name: %s", name)

        if not publisher:
            # TODO: Assumes "AmericanEnglish" publisher is the one that's used and wanted
            publisher = control_file_data[0x0200:0x020F].replace(b"\x00", b"").strip().decode("utf8")
        if not publisher:
            log.error("The Control Partition does not have any listed Publisher nor was one manually specified.")
            sys.exit(1)
        if not any(x in string.ascii_letters + string.digits for x in publisher):
            log.error(f'The Publisher, "{publisher}", cannot be all special characters.')
            sys.exit(1)
        if len(publisher.encode("utf8")) > 0x100:  # fits 0x100 * 10 (16 fields)
            log.error(f'The Title Publisher "{publisher}" is too large to fit in the NSP.')
            sys.exit(1)
        log.info("Publisher: %s", publisher)

        # only keep one name/publisher, store as AmericanEnglish
        # this is because the CLI has only one name/publisher option to use
        control_file_data[0x0:0x3000] = b"\x00" * 0x3000
        control_file_data[0x0:0x200] = name.encode("utf8").ljust(0x200, b"\x00")
        control_file_data[0x200:0x300] = publisher.encode("utf8").ljust(0x100, b"\x00")

        if version:
            version_utf8 = version.encode("utf8")
            while len(version_utf8) < 0x10:
                version_utf8 += b"\x00"
            control_file_data[0x3060:0x306F] = version_utf8
        else:
            version = control_file_data[0x3060:0x306F].replace(b"\x00", b"").strip().decode("utf8")
            if not version:
                log.error("The Control Partition does not have any listed Version nor was one manually specified.")
                sys.exit(1)
        if len(version.encode("utf8")) > 0x10:
            log.error(f'The Title Version "{version}" is too large to fit in the NSP.')
            sys.exit(1)
        log.info("Version: %s", version)

        control_file.write_bytes(control_file_data)

        if icon:
            shutil.copy(icon, icon_file)
        else:
            icon_file_res = nstool.get_icon(path, icon_file)
            if icon_file_res:
                if icon_file_res == "No Icon was extracted from the asset.":
                    log.warning("The NRO does not have an Icon, proceeding without one.")
                else:
                    log.critical(f"Failed extracting the Icon partition from the NRO, {icon_file_res}")
                    sys.exit(2)
            else:
                log.debug("Got the Icon partition")
                log.debug(base64.b64encode(icon_file.read_bytes()).decode())

        if icon_file.exists():
            # We must strip every unnecessary metadata or the icon will be a '?'
            im = Image.open(icon_file)
            if im.size != (256, 256):
                im = im.resize((256, 256))
            if im.mode != "RGB":
                im = im.convert("RGB")
            clean_im = Image.new(im.mode, im.size)
            clean_im.putdata(list(im.getdata()))
            clean_im.save(icon_file, format="JPEG")
            clean_im.close()
            im.close()

        next_nro_path = sdmc
        next_nro_path_file.write_text(next_nro_path)

        next_argv = next_nro_path
        if rom:
            next_argv += f' "sdmc:{rom}"'
        next_argv_file.write_text(next_argv)

        # only make this directory at this point because we have a high chance of success
        Directories.output.mkdir(parents=True, exist_ok=True)

        try:
            subprocess.check_output(
                [
                    Binaries.hacbrewpack,
                    "--titleid",
                    id_,
                    "--nspdir",
                    str(Directories.output.absolute()),
                    "-k",
                    str(Files.keys.absolute()),
                ],
                cwd=build_dir,
            )
            os.system("")  # fixes logs, I don't know why or how
        except subprocess.CalledProcessError as e:
            log.critical(f'Failed to build NSP, "{e.args}", {e.output} [{e.returncode}]')
            sys.exit(2)

        nsp_final_path = Directories.output / f"{name} v{version} by {publisher} [{id_}].nsp"
        if nsp_final_path.exists():
            log.warning("An NSP forwarder of the same name, publisher and title ID already existed.")
            nsp_final_path.unlink()
        shutil.move(Directories.output / f"{id_}.nsp", nsp_final_path)

        log.info(f"Done! The NSP has been saved to {nsp_final_path}")
    finally:
        if build_dir.exists():
            shutil.rmtree(build_dir)
        if hacbrewpack_backup_dir.exists():
            shutil.rmtree(hacbrewpack_backup_dir)


@main.command()
def update_game_ids() -> None:
    """
    Update the pre-existing Title ID registry.

    Note: This makes calls to Tinfoil.io that may fail.
    """
    print("Updating the Game Title ID registry via Tinfoil API, this may take a while...")
    game_title_ids = get_game_title_ids()
    Files.game_title_ids.write_text(jsonpickle.dumps(game_title_ids), encoding="utf8")
